
/* -*- Mode: Java; tab-width: 4; indent-tabs-mode: nil; c-basic-offset: 4 -*-
 *
 * This file is part of the IDE support for the AspectJ(tm)
 * programming language; see http://aspectj.org
 *
 * The contents of this file are subject to the Mozilla Public License
 * Version 1.1 (the "License"); you may not use this file except in
 * compliance with the License. You may obtain a copy of the License at
 * either http://www.mozilla.org/MPL/ or http://aspectj.org/MPL/.
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The Original Code is AspectJ.
 *
 * The Initial Developer of the Original Code is Xerox Corporation. Portions
 * created by Xerox Corporation are Copyright (C) 1999-2002 Xerox Corporation.
 * All Rights Reserved.
 *
 * Contributor(s):
 */

package org.aspectj.asm;

import java.util.*;
import java.io.*;

/**
 * Children are non-repeating making the parent-child structure a strict
 * tree.
 * 
 * @author Mik Kersten
 */
public abstract class StructureNode implements Serializable, Comparable {

	protected StructureNode parent = null;
    protected String name = "";
    protected String kind = "";
    protected List children = new ArrayList();
    protected StructureMessage message = null;
    protected SourceLocation sourceLocation = null;

    /**
     * Used during serialization.
     */
    public StructureNode() { }

    public StructureNode(String name, String kind, List children) {
        this.name = name;
        this.kind = kind;
        if (children != null) {
	        this.children = children;
        }
     	setParents();
    }

    public StructureNode(String name, String kind) {
        this.name = name;
        this.kind = kind;
    }

    public String toString() {
        return name;
    }

    public String getKind() {
        return kind;
    }

    public List getChildren() {
        return children;
    }

    public void addChild(StructureNode child) {
        if (children == null) {
            children = new ArrayList();
        }
        children.add(child);
        child.setParent(this);
    }
    
    public boolean removeChild(StructureNode child) {
    	child.setParent(null);
    	return children.remove(child);	
    }

	public StructureNode walk(ModelWalker walker) {
		for (Iterator it = children.iterator(); it.hasNext(); ) {
			StructureNode child = (StructureNode)it.next();
			walker.process(child);	
		} 
		return this;
	}
	
//	public boolean equals(Object o) {
//		if (!(o instanceof StructureNode)) return false;
//		StructureNode sn = (StructureNode)o;
//		return objectEqual(sn.getName(), this.getName())
//			&& objectEqual(sn.getKind(), this.getKind())
//			&& objectEqual(sn.getChildren(), this.getChildren());
//	}
//
//	protected boolean objectEqual(Object o1, Object o2) {
//		return (o1 == null && o2 == null) || (o1 != null && o1.equals(o2));		
//	}

    /**
     * Comparison is string-name based only.
     */
    public int compareTo(Object o) throws ClassCastException {
        if (this == o) {
            return 0;
        } else {
            StructureNode sn = (StructureNode)o;
            return this.getName().compareTo(sn.getName());
        }
    }

    public String getName() {
        return name;
    }

	public SourceLocation getSourceLocation() {
		return sourceLocation;
	}

	public void setSourceLocation(SourceLocation sourceLocation) {
		this.sourceLocation = sourceLocation;
	}

	public StructureMessage getMessage() {
		return message;
	}

	public void setMessage(StructureMessage message) {
		this.message = message;
	}

	public StructureNode getParent() {
		return parent;
	}

	public void setParent(StructureNode parent) {
		this.parent = parent;
	}

	private void setParents() {
		if (children == null) return;
		for (Iterator it = children.iterator(); it.hasNext(); ) {
			((StructureNode)it.next()).setParent(this);	
		}	
	}

//    /**
//     * Creates and returns a copy of this object.
//     */
//	public Object clone() {
//		List cloneChildren = new ArrayList();
//		for (Iterator it = children.iterator(); it.hasNext(); ) {
//			cloneChildren.add(((StructureNode)it.next()).clone());	
//		}
//    	StructureNode cloneNode = new StructureNode(name, kind, cloneChildren);
//    	return cloneNode;
//	} 
}


//    private void writeObject(ObjectOutputStream s) throws IOException {
//        s.defaultWriteObject();
//        // customized serialization code
//    }
//
//    private void readObject(ObjectInputStream s) throws IOException  {
//        s.defaultReadObject();
//        // customized deserialization code
//        ...
//        // followed by code to update the object, if necessary
//    }

//    public void writeExternal(ObjectOutput out) throws IOException {
//        if (this instanceof ProgramElementNode) {
//            out.writeInt(1);
//            writeString(name, out);
//            writeString(kind, out);
//            ((ProgramElementNode)this).writeExternal(out);
//        } if (this instanceof RelationNode) {
//            out.writeInt(1);
//            writeString(name, out);
//            writeString(kind, out);
//            ((RelationNode)this).writeExternal(out);
//        } if (this instanceof LinkNode) {
//            out.writeInt(3);
//            writeString(name, out);
//            writeString(kind, out);
//            ((LinkNode)this).writeExternal(out);
//        } else {
//            out.writeInt(0);
//            writeString(name, out);
//            writeString(kind, out);
//        }
//    }
//
//    public void readExternal(ObjectInput in) throws IOException, ClassNotFoundException {
//        int kindint = in.readInt();
//        name = readString(in);
//        kind = readString(in);
//
//        switch (kindint) {
//            case 1:
//                ((StructureNode)it.next()).readExternal(in);
//                break;
//            case 2:
//                ((RelationNode)it.next()).readExternal(in);
//                break;
//            case 3:
//                ((LinkNode)it.next()).readExternal(in);
//                break;
//        }
//    }
//
//    protected void writeString(String s, ObjectOutput out) throws IOException {
//        out.writeInt(s.length());
//        out.write(s.getBytes());
//    }
//
//    protected String readString(ObjectInput in) throws IOException {
//        int length = in.readInt();
//        byte[] nameArray = new byte[length];
//        in.read(nameArray, 0, length);
//        return new String(nameArray);
//    }